@@ -59,9 +59,10 @@
  * int
  * main (int argc, char **argv)
  * {
+ *  GtkWidget *mainwin;
  *   // Initialize i18n support with bindtextdomain(), etc.
  *
- *   ...
+ *   // ...
  *
  *   // Initialize the widget set
  *   gtk_init (&argc, &argv);
@@ -71,7 +72,7 @@
  *
  *   // Set up our GUI elements
  *
- *   ...
+ *   // ...
  *
  *   // Show the application window
  *   gtk_widget_show_all (mainwin);
@@ -91,6 +92,7 @@
 #include "config.h"
 
 #include "gdk/gdk.h"
+#include "gdk/gdk-private.h"
 
 #include <locale.h>
 
@@ -112,9 +114,9 @@
 
 #include "gtkaccelmapprivate.h"
 #include "gtkbox.h"
-#include "gtkclipboard.h"
+#include "gtkclipboardprivate.h"
 #include "gtkdebug.h"
-#include "gtkdnd.h"
+#include "gtkdndprivate.h"
 #include "gtkmain.h"
 #include "gtkmenu.h"
 #include "gtkmodules.h"
@@ -123,7 +125,7 @@
 #include "gtkrecentmanager.h"
 #include "gtkselectionprivate.h"
 #include "gtksettingsprivate.h"
-#include "gtktooltip.h"
+#include "gtktooltipprivate.h"
 #include "gtkversion.h"
 #include "gtkwidgetprivate.h"
 #include "gtkwindowprivate.h"
@@ -156,32 +158,39 @@ static GSList *main_loops = NULL;      /* stack of currently executing main loop
 
 static GSList *key_snoopers = NULL;
 
-static guint debug_flags = 0;              /* Global GTK debug flag */
+typedef struct {
+  GdkDisplay *display;
+  guint flags;
+} DisplayDebugFlags;
+
+#define N_DEBUG_DISPLAYS 4
+
+DisplayDebugFlags debug_flags[N_DEBUG_DISPLAYS];
 
-#ifdef G_ENABLE_DEBUG
 static const GDebugKey gtk_debug_keys[] = {
-  {"misc", GTK_DEBUG_MISC},
-  {"plugsocket", GTK_DEBUG_PLUGSOCKET},
-  {"text", GTK_DEBUG_TEXT},
-  {"tree", GTK_DEBUG_TREE},
-  {"updates", GTK_DEBUG_UPDATES},
-  {"keybindings", GTK_DEBUG_KEYBINDINGS},
-  {"multihead", GTK_DEBUG_MULTIHEAD},
-  {"modules", GTK_DEBUG_MODULES},
-  {"geometry", GTK_DEBUG_GEOMETRY},
-  {"icontheme", GTK_DEBUG_ICONTHEME},
-  {"printing", GTK_DEBUG_PRINTING},
-  {"builder", GTK_DEBUG_BUILDER},
-  {"size-request", GTK_DEBUG_SIZE_REQUEST},
-  {"no-css-cache", GTK_DEBUG_NO_CSS_CACHE},
-  {"baselines", GTK_DEBUG_BASELINES},
-  {"pixel-cache", GTK_DEBUG_PIXEL_CACHE},
-  {"no-pixel-cache", GTK_DEBUG_NO_PIXEL_CACHE},
-  {"interactive", GTK_DEBUG_INTERACTIVE},
-  {"touchscreen", GTK_DEBUG_TOUCHSCREEN},
-  {"actions", GTK_DEBUG_ACTIONS},
+  { "misc", GTK_DEBUG_MISC },
+  { "plugsocket", GTK_DEBUG_PLUGSOCKET },
+  { "text", GTK_DEBUG_TEXT },
+  { "tree", GTK_DEBUG_TREE },
+  { "updates", GTK_DEBUG_UPDATES },
+  { "keybindings", GTK_DEBUG_KEYBINDINGS },
+  { "multihead", GTK_DEBUG_MULTIHEAD },
+  { "modules", GTK_DEBUG_MODULES },
+  { "geometry", GTK_DEBUG_GEOMETRY },
+  { "icontheme", GTK_DEBUG_ICONTHEME },
+  { "printing", GTK_DEBUG_PRINTING} ,
+  { "builder", GTK_DEBUG_BUILDER },
+  { "size-request", GTK_DEBUG_SIZE_REQUEST },
+  { "no-css-cache", GTK_DEBUG_NO_CSS_CACHE },
+  { "baselines", GTK_DEBUG_BASELINES },
+  { "pixel-cache", GTK_DEBUG_PIXEL_CACHE },
+  { "no-pixel-cache", GTK_DEBUG_NO_PIXEL_CACHE },
+  { "interactive", GTK_DEBUG_INTERACTIVE },
+  { "touchscreen", GTK_DEBUG_TOUCHSCREEN },
+  { "actions", GTK_DEBUG_ACTIONS },
+  { "resize", GTK_DEBUG_RESIZE },
+  { "layout", GTK_DEBUG_LAYOUT }
 };
-#endif /* G_ENABLE_DEBUG */
 
 /**
  * gtk_get_major_version:
@@ -311,7 +320,7 @@ gtk_get_interface_age (void)
  * old version of gtk_check_version(), but still get loaded
  * into an application using a newer version of GTK+.
  *
- * Returns: %NULL if the GTK+ library is compatible with the
+ * Returns: (nullable): %NULL if the GTK+ library is compatible with the
  *   given version, or a string describing the version mismatch.
  *   The returned string is owned by GTK+ and should not be modified
  *   or freed.
@@ -414,9 +423,9 @@ static gboolean g_fatal_warnings = FALSE;
 static gboolean
 gtk_arg_debug_cb (const char *key, const char *value, gpointer user_data)
 {
-  debug_flags |= g_parse_debug_string (value,
-                                       gtk_debug_keys,
-                                       G_N_ELEMENTS (gtk_debug_keys));
+  debug_flags[0].flags |= g_parse_debug_string (value,
+                                                gtk_debug_keys,
+                                                G_N_ELEMENTS (gtk_debug_keys));
 
   return TRUE;
 }
@@ -424,9 +433,9 @@ gtk_arg_debug_cb (const char *key, const char *value, gpointer user_data)
 static gboolean
 gtk_arg_no_debug_cb (const char *key, const char *value, gpointer user_data)
 {
-  debug_flags &= ~g_parse_debug_string (value,
-                                        gtk_debug_keys,
-                                        G_N_ELEMENTS (gtk_debug_keys));
+  debug_flags[0].flags &= ~g_parse_debug_string (value,
+                                                 gtk_debug_keys,
+                                                 G_N_ELEMENTS (gtk_debug_keys));
 
   return TRUE;
 }
@@ -633,6 +642,7 @@ do_pre_parse_initialization (int    *argc,
                              char ***argv)
 {
   const gchar *env_string;
+  double slowdown;
   
   if (pre_initialized)
     return;
@@ -642,23 +652,46 @@ do_pre_parse_initialization (int    *argc,
   if (_gtk_module_has_mixed_deps (NULL))
     g_error ("GTK+ 2.x symbols detected. Using GTK+ 2.x and GTK+ 3 in the same process is not supported");
 
-  gdk_pre_parse_libgtk_only ();
+  GDK_PRIVATE_CALL (gdk_pre_parse) ();
   gdk_event_handler_set ((GdkEventFunc)gtk_main_do_event, NULL, NULL);
 
-#ifdef G_ENABLE_DEBUG
   env_string = g_getenv ("GTK_DEBUG");
   if (env_string != NULL)
     {
-      debug_flags = g_parse_debug_string (env_string,
-                                          gtk_debug_keys,
-                                          G_N_ELEMENTS (gtk_debug_keys));
+      debug_flags[0].flags = g_parse_debug_string (env_string,
+                                                   gtk_debug_keys,
+                                                   G_N_ELEMENTS (gtk_debug_keys));
+#ifndef G_ENABLE_DEBUG
+      /* No need to print the warning for "interactive" since it's kept anyway. */
+      if (debug_flags[0].flags != GTK_DEBUG_INTERACTIVE)
+        g_warning ("GTK_DEBUG set but ignored because gtk isn't built with G_ENABLE_DEBUG");
+      /* Only keep "interactive" if not with G_ENABLE_DEBUG. */
+      debug_flags[0].flags &= GTK_DEBUG_INTERACTIVE;
+#endif  /* G_ENABLE_DEBUG */
       env_string = NULL;
     }
-#endif  /* G_ENABLE_DEBUG */
+
+  env_string = g_getenv ("GTK3_MODULES");
+  if (env_string)
+    gtk_modules_string = g_string_new (env_string);
 
   env_string = g_getenv ("GTK_MODULES");
   if (env_string)
-    gtk_modules_string = g_string_new (env_string);
+    {
+      if (gtk_modules_string)
+        g_string_append_c (gtk_modules_string, G_SEARCHPATH_SEPARATOR);
+      else
+        gtk_modules_string = g_string_new (NULL);
+
+      g_string_append (gtk_modules_string, env_string);
+    }
+
+  env_string = g_getenv ("GTK_SLOWDOWN");
+  if (env_string)
+    {
+      slowdown = g_ascii_strtod (env_string, NULL);
+      _gtk_set_slowdown (slowdown);
+    }
 }
 
 static void
@@ -676,10 +709,19 @@ gettext_initialization (void)
 #endif  
 }
 
+static void
+default_display_notify_cb (GdkDisplayManager *dm)
+{
+  _gtk_accessibility_init ();
+  debug_flags[0].display = gdk_display_get_default ();
+}
+
 static void
 do_post_parse_initialization (int    *argc,
                               char ***argv)
 {
+  GdkDisplayManager *display_manager;
+
   if (gtk_initialized)
     return;
 
@@ -698,8 +740,10 @@ do_post_parse_initialization (int    *argc,
       g_log_set_always_fatal (fatal_mask);
     }
 
-  if (debug_flags & GTK_DEBUG_UPDATES)
+G_GNUC_BEGIN_IGNORE_DEPRECATIONS
+  if (debug_flags[0].flags & GTK_DEBUG_UPDATES)
     gdk_window_set_debug_updates (TRUE);
+G_GNUC_END_IGNORE_DEPRECATIONS
 
   gtk_widget_set_default_direction (gtk_get_locale_direction ());
 
@@ -707,11 +751,8 @@ do_post_parse_initialization (int    *argc,
 
   _gtk_accel_map_init ();
 
-  /* Set the 'initialized' flag.
-   */
   gtk_initialized = TRUE;
 
-  /* load gtk modules */
   if (gtk_modules_string)
     {
       _gtk_modules_init (argc, argv, gtk_modules_string->str);
@@ -722,7 +763,13 @@ do_post_parse_initialization (int    *argc,
       _gtk_modules_init (argc, argv, NULL);
     }
 
-  _gtk_accessibility_init ();
+  display_manager = gdk_display_manager_get ();
+  if (gdk_display_manager_get_default_display (display_manager) != NULL)
+    default_display_notify_cb (display_manager);
+
+  g_signal_connect (display_manager, "notify::default-display",
+                    G_CALLBACK (default_display_notify_cb),
+                    NULL);
 }
 
 
@@ -755,7 +802,7 @@ post_parse_hook (GOptionContext *context,
   
   if (info->open_default_display)
     {
-      if (gdk_display_open_default_libgtk_only () == NULL)
+      if (GDK_PRIVATE_CALL (gdk_display_open_default) () == NULL)
         {
           const char *display_name = gdk_get_display_arg_name ();
           g_set_error (error,
@@ -766,11 +813,46 @@ post_parse_hook (GOptionContext *context,
 
           return FALSE;
         }
+
+      if (gtk_get_debug_flags () & GTK_DEBUG_INTERACTIVE)
+        gtk_window_set_interactive_debugging (TRUE);
     }
 
   return TRUE;
 }
 
+guint
+gtk_get_display_debug_flags (GdkDisplay *display)
+{
+  gint i;
+
+  for (i = 0; i < N_DEBUG_DISPLAYS; i++)
+    {
+      if (debug_flags[i].display == display)
+        return debug_flags[i].flags;
+    }
+
+  return 0;
+}
+
+void
+gtk_set_display_debug_flags (GdkDisplay *display,
+                             guint       flags)
+{
+  gint i;
+
+  for (i = 0; i < N_DEBUG_DISPLAYS; i++)
+    {
+      if (debug_flags[i].display == NULL)
+        debug_flags[i].display = display;
+
+      if (debug_flags[i].display == display)
+        {
+          debug_flags[i].flags = flags;
+          return;
+        }
+    }
+}
 
 /**
  * gtk_get_debug_flags:
@@ -785,7 +867,7 @@ post_parse_hook (GOptionContext *context,
 guint
 gtk_get_debug_flags (void)
 {
-  return debug_flags;
+  return gtk_get_display_debug_flags (gdk_display_get_default ());
 }
 
 /**
@@ -796,11 +878,22 @@ gtk_get_debug_flags (void)
 void
 gtk_set_debug_flags (guint flags)
 {
-  debug_flags = flags;
+  gtk_set_display_debug_flags (gdk_display_get_default (), flags);
 }
 
+gboolean
+gtk_simulate_touchscreen (void)
+{
+  static gint test_touchscreen;
+
+  if (test_touchscreen == 0)
+    test_touchscreen = g_getenv ("GTK_TEST_TOUCHSCREEN") != NULL ? 1 : -1;
+
+  return test_touchscreen > 0 || (gtk_get_debug_flags () & GTK_DEBUG_TOUCHSCREEN) != 0;
+ }
+
 /**
- * gtk_get_option_group: (skip)
+ * gtk_get_option_group:
  * @open_default_display: whether to open the default display
  *     when parsing the commandline arguments
  *
@@ -830,7 +923,7 @@ gtk_get_option_group (gboolean open_default_display)
   group = g_option_group_new ("gtk", _("GTK+ Options"), _("Show GTK+ Options"), info, g_free);
   g_option_group_set_parse_hooks (group, pre_parse_hook, post_parse_hook);
 
-  gdk_add_option_entries_libgtk_only (group);
+  GDK_PRIVATE_CALL (gdk_add_option_entries) (group);
   g_option_group_add_entries (group, gtk_args);
   g_option_group_set_translation_domain (group, GETTEXT_PACKAGE);
   
@@ -850,7 +943,7 @@ gtk_get_option_group (gboolean open_default_display)
  *    `programname [OPTION...]`
  * @entries: (array zero-terminated=1): a %NULL-terminated array
  *    of #GOptionEntrys describing the options of your program
- * @translation_domain: a translation domain to use for translating
+ * @translation_domain: (nullable): a translation domain to use for translating
  *    the `--help` output for the options in @entries
  *    and the @parameter_string with gettext(), or %NULL
  * @error: a return location for errors
@@ -861,8 +954,9 @@ gtk_get_option_group (gboolean open_default_display)
  * `--help` output. Note that your program will
  * be terminated after writing out the help output.
  *
- * Returns: %TRUE if the windowing system has been successfully
- *     initialized, %FALSE otherwise
+ * Returns: %TRUE if the commandline arguments (if any) were valid and
+ *     if the windowing system has been successfully initialized,
+ *     %FALSE otherwise
  *
  * Since: 2.6
  */
@@ -879,14 +973,14 @@ gtk_init_with_args (gint                 *argc,
   gboolean retval;
 
   if (gtk_initialized)
-    return gdk_display_open_default_libgtk_only () != NULL;
+    goto done;
 
   gettext_initialization ();
 
   if (!check_setugid ())
     return FALSE;
 
-  gtk_group = gtk_get_option_group (TRUE);
+  gtk_group = gtk_get_option_group (FALSE);
 
   context = g_option_context_new (parameter_string);
   g_option_context_add_group (context, gtk_group);
@@ -898,7 +992,26 @@ gtk_init_with_args (gint                 *argc,
 
   g_option_context_free (context);
 
-  return retval;
+  if (!retval)
+    return FALSE;
+
+done:
+  if (GDK_PRIVATE_CALL (gdk_display_open_default) () == NULL)
+    {
+      const char *display_name = gdk_get_display_arg_name ();
+      g_set_error (error,
+                   G_OPTION_ERROR,
+                   G_OPTION_ERROR_FAILED,
+                   _("Cannot open display: %s"),
+                   display_name ? display_name : "" );
+
+      return FALSE;
+    }
+
+  if (gtk_get_debug_flags () & GTK_DEBUG_INTERACTIVE)
+    gtk_window_set_interactive_debugging (TRUE);
+
+  return TRUE;
 }
 
 
@@ -918,6 +1031,10 @@ gtk_init_with_args (gint                 *argc,
  * There is no need to call this function explicitly if you are using
  * gtk_init(), or gtk_init_check().
  *
+ * Note that many aspects of GTK+ require a display connection to
+ * function, so this way of initializing GTK+ is really only useful
+ * for specialized use cases.
+ *
  * Returns: %TRUE if initialization succeeded, otherwise %FALSE
  */
 gboolean
@@ -966,15 +1083,20 @@ gtk_parse_args (int    *argc,
  *     understood by GTK+ are stripped before return.
  *
  * This function does the same work as gtk_init() with only a single
- * change: It does not terminate the program if the windowing system
- * can’t be initialized. Instead it returns %FALSE on failure.
+ * change: It does not terminate the program if the commandline
+ * arguments couldn’t be parsed or the windowing system can’t be
+ * initialized. Instead it returns %FALSE on failure.
  *
  * This way the application can fall back to some other means of
  * communication with the user - for example a curses or command line
  * interface.
  *
- * Returns: %TRUE if the windowing system has been successfully
- *     initialized, %FALSE otherwise
+ * Note that calling any GTK function or instantiating any GTK type after
+ * this function returns %FALSE results in undefined behavior.
+ *
+ * Returns: %TRUE if the commandline arguments (if any) were valid and
+ *     the windowing system has been successfully initialized, %FALSE
+ *     otherwise
  */
 gboolean
 gtk_init_check (int    *argc,
@@ -985,9 +1107,9 @@ gtk_init_check (int    *argc,
   if (!gtk_parse_args (argc, argv))
     return FALSE;
 
-  ret = gdk_display_open_default_libgtk_only () != NULL;
+  ret = GDK_PRIVATE_CALL (gdk_display_open_default) () != NULL;
 
-  if (debug_flags & GTK_DEBUG_INTERACTIVE)
+  if (gtk_get_debug_flags () & GTK_DEBUG_INTERACTIVE)
     gtk_window_set_interactive_debugging (TRUE);
 
   return ret;
@@ -1157,7 +1279,7 @@ gtk_get_locale_direction (void)
   if (g_strcmp0 (e, "default:RTL") == 0)
     dir = GTK_TEXT_DIR_RTL;
   else if (g_strcmp0 (e, "default:LTR") != 0)
-    g_warning ("Whoever translated default:LTR did so wrongly. Defaulting to LTR.\n");
+    g_warning ("Whoever translated default:LTR did so wrongly. Defaulting to LTR.");
 
   return dir;
 }
@@ -1206,7 +1328,10 @@ gtk_main (void)
       gdk_threads_leave ();
       g_main_loop_run (loop);
       gdk_threads_enter ();
+
+      G_GNUC_BEGIN_IGNORE_DEPRECATIONS;
       gdk_flush ();
+      G_GNUC_END_IGNORE_DEPRECATIONS;
     }
 
   main_loops = g_slist_remove (main_loops, loop);
@@ -1338,12 +1463,6 @@ gtk_main_iteration_do (gboolean blocking)
     return TRUE;
 }
 
-/* private libgtk to libgdk interfaces */
-gboolean gdk_device_grab_info_libgtk_only (GdkDisplay  *display,
-                                           GdkDevice   *device,
-                                           GdkWindow  **grab_window,
-                                           gboolean    *owner_events);
-
 static void
 rewrite_events_translate (GdkWindow *old_window,
                           GdkWindow *new_window,
@@ -1394,6 +1513,18 @@ rewrite_event_for_window (GdkEvent  *event,
                                 new_window,
                                 &event->touch.x, &event->touch.y);
       break;
+    case GDK_TOUCHPAD_SWIPE:
+      rewrite_events_translate (event->any.window,
+                                new_window,
+                                &event->touchpad_swipe.x,
+                                &event->touchpad_swipe.y);
+      break;
+    case GDK_TOUCHPAD_PINCH:
+      rewrite_events_translate (event->any.window,
+                                new_window,
+                                &event->touchpad_pinch.x,
+                                &event->touchpad_pinch.y);
+      break;
     case GDK_KEY_PRESS:
     case GDK_KEY_RELEASE:
     case GDK_PROXIMITY_IN:
@@ -1443,10 +1574,12 @@ rewrite_event_for_grabs (GdkEvent *event)
     case GDK_TOUCH_UPDATE:
     case GDK_TOUCH_END:
     case GDK_TOUCH_CANCEL:
+    case GDK_TOUCHPAD_SWIPE:
+    case GDK_TOUCHPAD_PINCH:
       display = gdk_window_get_display (event->any.window);
       device = gdk_event_get_device (event);
 
-      if (!gdk_device_grab_info_libgtk_only (display, device, &grab_window, &owner_events) ||
+      if (!GDK_PRIVATE_CALL (gdk_device_grab_info) (display, device, &grab_window, &owner_events) ||
           !owner_events)
         return NULL;
       break;
@@ -1465,6 +1598,54 @@ rewrite_event_for_grabs (GdkEvent *event)
     return NULL;
 }
 
+static GtkWidget *
+widget_get_popover_ancestor (GtkWidget *widget,
+                             GtkWindow *window)
+{
+  GtkWidget *parent = gtk_widget_get_parent (widget);
+
+  while (parent && parent != GTK_WIDGET (window))
+    {
+      widget = parent;
+      parent = gtk_widget_get_parent (widget);
+    }
+
+  if (!parent || parent != GTK_WIDGET (window))
+    return NULL;
+
+  if (_gtk_window_is_popover_widget (GTK_WINDOW (window), widget))
+    return widget;
+
+  return NULL;
+}
+
+static gboolean
+check_event_in_child_popover (GtkWidget *event_widget,
+                              GtkWidget *grab_widget)
+{
+  GtkWidget *window, *popover = NULL, *popover_parent = NULL;
+
+  if (grab_widget == event_widget)
+    return FALSE;
+
+  window = gtk_widget_get_ancestor (event_widget, GTK_TYPE_WINDOW);
+
+  if (!window)
+    return FALSE;
+
+  popover = widget_get_popover_ancestor (event_widget, GTK_WINDOW (window));
+
+  if (!popover)
+    return FALSE;
+
+  popover_parent = _gtk_window_get_popover_parent (GTK_WINDOW (window), popover);
+
+  if (!popover_parent)
+    return FALSE;
+
+  return (popover_parent == grab_widget || gtk_widget_is_ancestor (popover_parent, grab_widget));
+}
+
 /**
  * gtk_main_do_event:
  * @event: An event to process (normally passed by GDK)
@@ -1561,11 +1742,10 @@ gtk_main_do_event (GdkEvent *event)
       event_widget = gtk_get_event_widget (event);
     }
 
-  if (GTK_IS_WINDOW (event_widget))
-    {
-      if (_gtk_window_check_handle_wm_event (event))
-        return;
-    }
+  /* Push the event onto a stack of current events for
+   * gtk_current_event_get().
+   */
+  current_events = g_list_prepend (current_events, event);
 
   window_group = gtk_main_get_window_group (event_widget);
   device = gdk_event_get_device (event);
@@ -1577,6 +1757,19 @@ gtk_main_do_event (GdkEvent *event)
   if (!grab_widget)
     grab_widget = gtk_window_group_get_current_grab (window_group);
 
+  if (GTK_IS_WINDOW (event_widget) ||
+      (grab_widget && grab_widget != event_widget &&
+       !gtk_widget_is_ancestor (event_widget, grab_widget)))
+    {
+      /* Ignore event if we got a grab on another toplevel */
+      if (!grab_widget ||
+          gtk_widget_get_toplevel (event_widget) == gtk_widget_get_toplevel (grab_widget))
+        {
+          if (_gtk_window_check_handle_wm_event (event))
+            goto cleanup;
+        }
+    }
+
   /* Find out the topmost widget where captured event propagation
    * should start, which is the widget holding the GTK+ grab
    * if any, otherwise it's left NULL and events are emitted
@@ -1594,22 +1787,20 @@ gtk_main_do_event (GdkEvent *event)
        gtk_widget_is_ancestor (event_widget, grab_widget)))
     grab_widget = event_widget;
 
+  /* popovers are not really a "child" of their "parent" in the widget/window
+   * hierarchy sense, we however want to interact with popovers spawn by widgets
+   * within grab_widget. If this is the case, we let the event go through
+   * unaffected by the grab.
+   */
+  if (check_event_in_child_popover (event_widget, grab_widget))
+    grab_widget = event_widget;
+
   /* If the widget receiving events is actually blocked by another
    * device GTK+ grab
    */
   if (device &&
       _gtk_window_group_widget_is_blocked_for_device (window_group, grab_widget, device))
-    {
-      if (rewritten_event)
-        gdk_event_free (rewritten_event);
-
-      return;
-    }
-
-  /* Push the event onto a stack of current events for
-   * gtk_current_event_get().
-   */
-  current_events = g_list_prepend (current_events, event);
+    goto cleanup;
 
   /* Not all events get sent to the grabbing widget.
    * The delete, destroy, expose, focus change and resize
@@ -1650,32 +1841,7 @@ gtk_main_do_event (GdkEvent *event)
 
     case GDK_EXPOSE:
       if (event->any.window)
-        {
-          gboolean is_double_buffered;
-
-          G_GNUC_BEGIN_IGNORE_DEPRECATIONS;
-          is_double_buffered = gtk_widget_get_double_buffered (event_widget);
-          G_GNUC_END_IGNORE_DEPRECATIONS;
-
-          if (is_double_buffered)
-            {
-              /* We handle exposes only on native windows, relying on the
-               * draw() handler to propagate down to non-native windows.
-               * This is ok now that we child windows always are considered
-               * (semi)transparent.
-               */
-              if (gdk_window_has_native (event->expose.window))
-                {
-                  gdk_window_begin_paint_region (event->any.window, event->expose.region);
-                  gtk_widget_send_expose (event_widget, event);
-                  gdk_window_end_paint (event->any.window);
-                }
-            }
-          else
-            {
-              gtk_widget_send_expose (event_widget, event);
-            }
-        }
+        gtk_widget_render (event_widget, event->any.window, event->expose.region);
       break;
 
     case GDK_PROPERTY_NOTIFY:
@@ -1744,6 +1910,13 @@ gtk_main_do_event (GdkEvent *event)
     case GDK_TOUCH_UPDATE:
     case GDK_TOUCH_END:
     case GDK_TOUCH_CANCEL:
+    case GDK_TOUCHPAD_SWIPE:
+    case GDK_TOUCHPAD_PINCH:
+    case GDK_PAD_BUTTON_PRESS:
+    case GDK_PAD_BUTTON_RELEASE:
+    case GDK_PAD_RING:
+    case GDK_PAD_STRIP:
+    case GDK_PAD_GROUP_MODE:
       if (!_gtk_propagate_captured_event (grab_widget, event, topmost_widget))
         gtk_propagate_event (grab_widget, event);
       break;
@@ -1785,6 +1958,7 @@ gtk_main_do_event (GdkEvent *event)
       _gtk_tooltip_handle_event (event);
     }
 
+ cleanup:
   tmp_list = current_events;
   current_events = g_list_remove_link (current_events, tmp_list);
   g_list_free_1 (tmp_list);
@@ -2083,7 +2257,7 @@ gtk_grab_add (GtkWidget *widget)
  *
  * Queries the current grab of the default window group.
  *
- * Returns: (transfer none): The widget which currently
+ * Returns: (transfer none) (nullable): The widget which currently
  *     has the grab or %NULL if no grab is active
  */
 GtkWidget*
@@ -2285,7 +2459,7 @@ gtk_invoke_key_snoopers (GtkWidget *grab_widget,
  * the current event will be the #GdkEventButton that triggered
  * the ::clicked signal.
  *
- * Returns: (transfer full): a copy of the current event, or
+ * Returns: (transfer full) (nullable): a copy of the current event, or
  *     %NULL if there is no current event. The returned event must be
  *     freed with gdk_event_free().
  */
@@ -2347,7 +2521,7 @@ gtk_get_current_event_state (GdkModifierType *state)
  * If there is a current event and it has a device, return that
  * device, otherwise return %NULL.
  *
- * Returns: (transfer none): a #GdkDevice, or %NULL
+ * Returns: (transfer none) (nullable): a #GdkDevice, or %NULL
  */
 GdkDevice *
 gtk_get_current_event_device (void)
@@ -2366,7 +2540,7 @@ gtk_get_current_event_device (void)
  * returns %NULL, otherwise returns the widget that received the event
  * originally.
  *
- * Returns: (transfer none): the widget that originally
+ * Returns: (transfer none) (nullable): the widget that originally
  *     received @event, or %NULL
  */
 GtkWidget*
@@ -2450,7 +2624,7 @@ propagate_event_down (GtkWidget *widget,
         break;
     }
 
-  for (l = widgets; l && !handled_event; l = g_list_next (l))
+  for (l = widgets; l && !handled_event; l = l->next)
     {
       widget = (GtkWidget *)l->data;
 
